"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import inspect as nspt
import typing as h
from importlib import util as mprt
from pathlib import Path as pl_path_t

from babelwidget.api.backend import backend_t
from babelwidget.api.widget import base_h as library_wgt_h
from conf_ini_g.catalog.choices import choices_wgt_t
from conf_ini_g.catalog.path import path_wgt_t
from conf_ini_g.interface.window.parameter.value import value_wgt_a
from value_factory.api.catalog import callable_t, path_kind_e, path_purpose_e, path_t
from value_factory.api.type import hint_t


@d.dataclass(repr=False, eq=False)
class callable_wgt_t(value_wgt_a):
    library_wgt: library_wgt_h
    kind: h.Literal["class", "function"]
    default_choices: choices_wgt_t | None = d.field(init=False, default=None)
    path: path_wgt_t | None = d.field(init=False, default=None)
    choices: choices_wgt_t | None = d.field(init=False, default=None)

    @classmethod
    def New(
        cls,
        stripe: hint_t,
        backend: backend_t,
        /,
    ) -> h.Self:
        """"""
        if (annotation := stripe.FirstAnnotationWithType(callable_t)) is None:
            kind = "function"
            default_choices = None
            allow_external = True
        else:
            kind = annotation.kind
            default_choices = annotation.choices
            allow_external = annotation.allow_external

        assert (default_choices is not None) or allow_external

        if default_choices is None:
            default_choices_wgt = None
        else:
            default_choices_wgt = choices_wgt_t.New(default_choices, backend)

        # TODO: Check emitter and signal.
        output = cls(
            default_choices_wgt,
            "currentIndexChanged",
            backend,
            library_wgt=backend.base_t(),
            kind=kind,
        )
        output.default_choices = default_choices_wgt

        kind_wgt = backend.label_t(kind)

        layout = backend.hbox_lyt_t()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(kind_wgt)
        if allow_external:
            hint = hint_t.New(
                path_t(kind=path_kind_e.document, purpose=path_purpose_e.input)
            )
            path_wgt = path_wgt_t.New(
                hint,
                backend,
                PostAssignmentFunction=output._UpdateChoices,
            )
            choices_wgt = choices_wgt_t.New(None, backend)
            output.path = path_wgt
            output.choices = choices_wgt

            if default_choices_wgt is not None:
                layout.addWidget(default_choices_wgt.library_wgt)
            layout.addWidget(path_wgt.library_wgt)
            layout.addWidget(choices_wgt.library_wgt)
        else:
            layout.addWidget(default_choices_wgt.library_wgt)
        output.library_wgt.setLayout(layout)

        return output

    def _UpdateChoices(self, path: pl_path_t, /) -> None:
        """"""
        self.choices.library_wgt.clear()

        spec = mprt.spec_from_file_location(path.stem, path)
        module = spec.loader.load_module(spec.name)
        if self.kind == "class":
            condition = nspt.isclass
        else:
            condition = nspt.isfunction

        for name, _ in nspt.getmembers(module, condition):
            if not name.startswith("_"):
                self.choices.library_wgt.addItem(name)

    def SetFunction(self, function: h.Callable, /) -> None:
        """
        self.activated.connect(function): Responds only to GUI interaction, not
        programmatic change.
        """
        if self.default_choices is not None:
            self.default_choices.currentIndexChanged.connect(function)
        if self.choices is not None:
            self.choices.currentIndexChanged.connect(function)

    def Assign(self, value: str, _: h.Any, /) -> None:
        """"""
        if callable_t.SEPARATOR in value:
            path, choice = value.split(callable_t.SEPARATOR)
            self.path.Assign(pl_path_t(path), None)
            self.choices.Assign(choice, None)
        else:
            self.default_choices.Assign(value, None)

    def Text(self) -> str:
        """"""
        if self.path is None:
            return self.default_choices.Text()

        path = self.path.Text()
        if path.__len__() > 0:
            return f"{path}{callable_t.SEPARATOR}{self.choices.Text()}"

        if self.default_choices is None:
            return ""

        return self.default_choices.Text()


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import typing as h

from babelwidget.api.backend import backend_t
from babelwidget.api.widget import grid_lyt_h
from babelwidget.api.widget import group_h as group_wgt_h
from babelwidget.api.widget import label_h as label_wgt_h
from babelwidget.api.widget import stack_h as stack_wgt_h
from conf_ini_g.constant.section import INI_UNIT_SECTION
from conf_ini_g.interface.window.parameter.main import parameter_t as visual_parameter_t
from conf_ini_g.task.formatting import FormattedName
from conf_ini_g.type.dict import section_str_h
from conf_ini_g.type.parameter import prm_t as functional_parameter_t
from conf_ini_g.type.section import section_controlled_t as functional_controlled_t
from conf_ini_g.type.section import section_free_t as functional_free_t


@d.dataclass(repr=False, eq=False)
class _base_t:  # Cannot be abstracted.
    HEADER_NAMES: h.ClassVar[tuple[str]] = (
        "Parameter",
        "Type(s)",
        "Value",
        "Unit",
    )
    HEADER_STYLE: h.ClassVar[str] = "background-color: darkgray; padding-left: 5px;"

    functional: functional_free_t | functional_controlled_t
    library_wgt: group_wgt_h

    def __post_init__(self) -> None:
        """"""
        self.functional.interface = self

    @classmethod
    def NewWithName(
        cls,
        name: str,
        functional: functional_free_t | functional_controlled_t,
        backend: backend_t,
        /,
    ) -> h.Self:
        """"""
        output = cls(functional=functional, library_wgt=backend.group_t())

        if functional.__class__.is_controlled:
            controller = (
                f" ← {FormattedName(functional.controller[0])}."
                f"{FormattedName(functional.controller[1])}"
            )
        else:
            controller = ""
        formatted = FormattedName(name) + controller
        output.library_wgt.setTitle(formatted)

        return output

    @classmethod
    def Headers(cls, backend: backend_t, /) -> list[label_wgt_h]:
        """"""
        output = []

        for text in cls.HEADER_NAMES:
            header = backend.label_t(f'<font color="blue">{text}</font>')
            header.setStyleSheet(cls.HEADER_STYLE)
            output.append(header)

        return output


@d.dataclass(repr=False, eq=False)
class section_free_t(_base_t):

    @property
    def active_as_str_dict(self) -> section_str_h:
        """"""
        return {_key: _vle.interface.Text() for _key, _vle in self.functional.items()}

    @classmethod
    def New(
        cls,
        name: str,
        functional: functional_free_t,
        backend: backend_t,
        /,
    ) -> h.Self | None:
        """"""
        output = cls.NewWithName(name, functional, backend)

        parameters, layout = _VisualParameters(
            tuple(functional.prm_iterator), name == INI_UNIT_SECTION, backend
        )
        if parameters.__len__() == 0:
            return None

        for h_idx, header in enumerate(cls.Headers(backend)):
            layout.addWidget(header, 0, h_idx)
        output.library_wgt.setLayout(layout)

        return output


@d.dataclass(repr=False, eq=False)
class section_controlled_t(_base_t):
    subset_stack: stack_wgt_h = d.field(init=False)

    @property
    def active_as_str_dict(self) -> section_str_h:
        """"""
        output = {}

        controlling_value = self.functional.controller[2].interface.Text()
        if None in self.functional:
            values = (None, controlling_value)
        else:
            values = (controlling_value,)
        for value in values:
            output.update(
                {
                    _key: _vle.interface.Text()
                    for _key, _vle in self.functional[value].items()
                }
            )

        return output

    @classmethod
    def New(
        cls,
        name: str,
        functional: functional_controlled_t,
        backend: backend_t,
        /,
    ) -> h.Self | None:
        """"""
        output = cls.NewWithName(name, functional, backend)

        subset_stack = backend.stack_t()
        for (
            value,
            functional_s,
        ) in functional.prm_iterator_per_controlling_value:
            parameters, layout = _VisualParameters(functional_s, False, backend)
            if parameters.__len__() == 0:
                continue

            for h_idx, header in enumerate(cls.Headers(backend)):
                layout.addWidget(header, 0, h_idx)
            page = backend.base_t()
            page.setLayout(layout)
            subset_stack.addWidget(page)

        output.subset_stack = subset_stack

        # Curiously, the stacked widget cannot be simply declared as child of instance;
        # This must be specified through a layout.
        layout = backend.hbox_lyt_t()
        layout.addWidget(subset_stack)
        layout.setContentsMargins(0, 0, 0, 0)
        output.library_wgt.setLayout(layout)

        return output


any_section_h = section_free_t | section_controlled_t


def _VisualParameters(
    functional_s: h.Sequence[tuple[str, functional_parameter_t]],
    section_is_unit: bool,
    backend: backend_t,
    /,
) -> tuple[list[visual_parameter_t], grid_lyt_h]:
    """"""
    out_prm_s = []

    out_lyt = backend.grid_lyt_t()
    out_lyt.setAlignment(backend.ALIGNED_TOP)
    out_lyt.setColumnStretch(0, 4)
    out_lyt.setColumnStretch(1, 1)
    out_lyt.setColumnStretch(2, 8)
    out_lyt.setColumnStretch(3, 1)
    out_lyt.setContentsMargins(0, 0, 0, 0)

    for row, (name, functional) in enumerate(functional_s, start=1):
        parameter = visual_parameter_t.New(name, functional, backend)
        out_prm_s.append(parameter)

        out_lyt.addWidget(parameter.name, row, 0, alignment=backend.ALIGNED_RIGHT)
        out_lyt.addWidget(parameter.type, row, 1)
        out_lyt.addWidget(parameter.value.library_wgt, row, 2, 1, 2 - 1)
        if not section_is_unit:
            out_lyt.addWidget(parameter.unit, row, 3)

    return out_prm_s, out_lyt


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import textwrap as text
import typing as h

import scipy.constants as unit
from conf_ini_g.constant.console import MARGIN
from conf_ini_g.task.conversion import ValueInStandardUnit
from conf_ini_g.type.dict import section_name_h
from conf_ini_g.type.property import ppt_t
from logger_36 import L
from value_factory.api.catalog import annotation_a
from value_factory.api.constant import UNSET
from value_factory.api.extension import hint_h
from value_factory.api.task import NewCastValue
from value_factory.api.type import (
    hint_t,
    invalid_t,
    unset_t,
    value_passed_h,
    value_wanted_h,
)

MustRespect_h = h.Callable[[value_wanted_h], bool]


@d.dataclass(slots=True, repr=False, eq=False)
class prm_t:
    hint: hint_h | hint_t | annotation_a = h.Any
    unit_name: str | None = None
    unit_factor: float | None = None
    default: value_wanted_h = UNSET
    _value: value_wanted_h = d.field(init=False, default=UNSET)
    must: MustRespect_h | tuple[MustRespect_h, ...] | None = None
    ppt: ppt_t | None = None
    interface: h.Any = None

    @property
    def value(self) -> value_wanted_h | invalid_t | unset_t:
        """"""
        if self._value is UNSET:
            if self.default is UNSET:
                return UNSET
            output = self.default
        else:
            output = self._value

        if self.unit_factor is None:
            return output

        return ValueInStandardUnit(output, self.unit_name, self.unit_factor)

    def __post_init__(self) -> None:
        """"""
        assert not isinstance(self.default, invalid_t)

        if not isinstance(self.hint, hint_t):
            self.hint = hint_t.New(self.hint)

        if self.must is None:
            self.must = ()
        elif isinstance(self.must, h.Callable):
            self.must = (self.must,)
        elif not isinstance(self.must, tuple):
            self.must = tuple(self.must)

        if self.ppt is None:
            self.ppt = ppt_t()
        self.ppt.optional = self.default is not UNSET

        self.Issues()

    @classmethod
    def NewRuntime(cls, value: h.Any, basic: bool, /) -> h.Self:
        """
        Runtime means "not part of the configuration definition".
        """
        output = cls(
            hint=type(value),
            ppt=ppt_t(
                short="Runtime parameter",
                long="This parameter is not part of the configuration definition. "
                "It was encountered in an INI configuration and added programmatically.",
                basic=basic,
            ),
        )
        output._value = value
        return output

    def Issues(
        self,
        /,
        *,
        should_raise: bool = False,
        should_be_staged: bool = False,
        can_be_none: bool = False,
    ) -> list[str] | None | int | h.NoReturn:
        """"""
        output = []

        if self.ppt.optional:
            converted, issues = NewCastValue(self.default, self.hint)
            if issues.__len__() > 0:
                output.extend(issues)
            else:
                self.default = converted
                for must in self.must:
                    if not must(self.default):
                        output.append(f"{must.__name__}({self.default}) is False.")
        elif not self.ppt.basic:
            output.append("Parameter is not basic but not optional.")

        if (n_output := output.__len__()) > 0:
            if should_raise:
                raise Exception("\n".join(output))
            if should_be_staged:
                for element in output:
                    L.StageIssue(element)
                return n_output
            return output

        if should_raise:
            return None
        if should_be_staged:
            return 0
        if can_be_none:
            return None
        return []

    def Assign(self, value: value_passed_h | invalid_t | unset_t, /) -> list[str]:
        """"""
        if isinstance(value, invalid_t | unset_t):
            self._value = value
            return [f"Unexpected value {value}."]

        converted, issues = NewCastValue(value, self.hint)
        if issues.__len__() > 0:
            return issues

        does_not_respect = ", ".join(
            getattr(_, "__name__", str(_)) for _ in self.must if not _(converted)
        )
        if does_not_respect.__len__() > 0:
            return [f"{converted} does not respect {does_not_respect}."]

        self._value = converted

        return []

    def SetUnit(self, passed: str | int | float, /) -> list[str]:
        """"""
        if isinstance(passed, int | float):
            self.unit_name = f"<unit*{passed}>"
            self.unit_factor = passed
            return []

        unit_as_float = getattr(unit, passed, None)
        if unit_as_float is None:
            return [
                f"Unit {passed} is not defined by the SciPy library. "
                f"See: https://docs.scipy.org/doc/scipy/reference/constants.html"
            ]

        self.unit_name = passed
        self.unit_factor = unit_as_float
        return []

    def __str__(self) -> str:
        """"""
        type_options = self.hint
        if isinstance(type_options, type):
            type_options = type_options.__name__

        if self._value is UNSET:
            if self.default is UNSET:
                value = ""
            else:
                value = (
                    f"Value: {self.default}:{type(self.default).__name__} (default)\n"
                )
        else:
            if self.unit_factor is None:
                unit_name = ""
            else:
                unit_name = f" {self.unit_name}"
            value = f"Value: {self._value}{unit_name}:{type(self._value).__name__}\n"

        must = ", ".join(getattr(_, "__name__", str(_)) for _ in self.must)
        return (
            f"Properties:\n{text.indent(str(self.ppt), MARGIN)}\n"
            f"Type options: {type_options}\n"
            f"{value}"
            f"Must respect: {must}"
        )

    __repr__ = __str__


@d.dataclass(slots=True, repr=False, eq=False)
class ctl_t(prm_t):
    controlled_section: section_name_h = ""

    def __post_init__(self) -> None:
        """
        A controller must not have an explicit default value; The default value is set
        as the first non-None key of the controlled section.
        """
        prm_t.__post_init__(self)

        assert self.hint.type is h.Any
        assert (self.unit_name is None) and (self.unit_factor is None)
        assert self.default is UNSET
        assert self.must == ()
        assert not self.ppt.optional
        assert isinstance(self.controlled_section, str) and (
            self.controlled_section.__len__() > 0
        )

        self.hint.type = str

    def AddControllingValue(self, value: str, /) -> None:
        """"""
        assert isinstance(value, str), value

        if self.hint.is_literal:
            assert self.hint.is_str_literal, self.hint
            self.hint.AddLiteral(value)
        else:
            assert self.hint.type is str, self.hint
            self.hint = hint_t.New(h.Literal[(value,)])

        if self._value is UNSET:
            self._value = value

    def SetControllingValues(self, values: tuple[str, ...], /) -> None:
        """"""
        values = tuple(sorted(values))
        self.hint = hint_t.New(h.Literal[values])
        self._value = values[0]

    def __str__(self) -> str:
        """"""
        return f"{prm_t.__str__(self)}\nControls: {self.controlled_section}"

    __repr__ = __str__


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import textwrap as text
import typing as h
from pathlib import Path as path_t

from conf_ini_g.constant.console import MARGIN
from conf_ini_g.constant.section import INI_UNIT_SECTION
from conf_ini_g.interface.storage.config import INIConfig
from conf_ini_g.type.dict import config_str_h, config_typed_h, section_name_h
from conf_ini_g.type.parameter import ctl_t, prm_t
from conf_ini_g.type.property import ppt_t
from conf_ini_g.type.section import (
    section_controlled_t,
    section_definition_h,
    section_free_t,
)
from logger_36 import L
from value_factory.api.constant import ATOMIC_FACTORIES, UNSET
from value_factory.extension.hint import hint_h

config_definition_h = dict[section_name_h, section_definition_h]

_SECTION_TYPE = {False: section_free_t, True: section_controlled_t}


@d.dataclass(slots=True, repr=False, eq=False)
class config_t(dict[section_name_h, section_free_t | section_controlled_t]):

    definition: config_definition_h
    path: path_t | None = d.field(init=False, default=None)
    interface: h.Any = None

    @property
    def active_as_typed_dict(self) -> config_typed_h:
        """"""
        return {_snm: _sct.active_as_typed_dict for _snm, _sct in self.items()}

    def __post_init__(self) -> None:
        """"""
        controlled_sections = []
        for section in self.definition.values():
            for content in section.values():
                if isinstance(content, ctl_t):
                    controlled_sections.append(content.controlled_section)

        for name, section in self.definition.items():
            assert name.isidentifier(), name
            self[name] = _SECTION_TYPE[name in controlled_sections].New(section)

        for s_name, section in self.items():
            for p_name, content in section.items():
                if not isinstance(content, ctl_t):
                    continue

                target = self[content.controlled_section]  # Can "except".
                target.controller = (s_name, p_name, content)
                if (values := target.controlling_values) is not UNSET:
                    content.SetControllingValues(values)

        for section in self.values():
            if section.__class__.is_controlled:
                assert isinstance(section.controller, tuple), section.controller
            section.Issues(should_raise=True)

        # Wait for the config to be fully populated (optional plugin additions) before
        # checking if it has issues.

    def AddPluginParameter(
        self,
        s_name: str,
        p_name: str,
        /,
        *,
        hint: hint_h | None = None,
        default: h.Any = UNSET,
        controlling_value: str | None = None,
        short: str = "Plugin parameter",
        long: str = "This parameter is not part of the configuration definition. "
        'It was added programmatically in a "plugin" way.',
        basic: bool = True,
    ) -> None:
        """"""
        assert s_name.isidentifier(), s_name
        assert p_name.isidentifier(), p_name

        target = self[s_name]

        if not target.ppt.growable:
            raise RuntimeError(
                f"{s_name}.{p_name}: "
                f"Attempt to add an unspecified parameter to a section accepting none."
            )

        if not (target.__class__.is_controlled or (controlling_value is None)):
            raise RuntimeError(
                f"{s_name}.{p_name}<={controlling_value}: "
                f"Attempt to add a controlled parameter to an uncontrolled section."
            )

        if target.__len__() > 0:
            if target.__class__.is_controlled:
                error = p_name in target.prm_free_names
                if not (
                    error
                    or (controlling_value is None)
                    or (controlling_value not in target.controlling_values)
                ):
                    error = p_name in target[controlling_value]
            else:
                error = p_name in target
            if error:
                raise RuntimeError(
                    f"A parameter {p_name} already exists in section {s_name}."
                )

        parameter = prm_t(
            hint=hint,
            default=default,
            ppt=ppt_t(short=short, long=long, basic=basic),
        )

        if target.__class__.is_controlled:
            if controlling_value is None:
                if None in target:
                    target[None][p_name] = parameter
                else:
                    target[None] = {p_name: parameter}
            elif ((values := target.controlling_values) is not UNSET) and (
                controlling_value in values
            ):
                target[controlling_value][p_name] = parameter
            else:
                target[controlling_value] = {p_name: parameter}
                target.controller[2].AddControllingValue(controlling_value)
        else:
            target[p_name] = parameter

    def UpdateFromINI(self, path: str | path_t, /) -> list[str]:
        """"""
        ini_config = INIConfig(path)
        issues = self.UpdateFromDict(ini_config, should_interpret_values=True)
        if issues.__len__() > 0:
            self.path = None
            return issues

        self.path = path_t(path)
        return []

    def UpdateFromDict(
        self,
        config: config_str_h | config_typed_h,
        /,
        *,
        should_interpret_values: bool = True,
        should_update_interface: bool = True,
    ) -> list[str]:
        """
        should_interpret_values: Applies only to parameters not in the definition inside
        growable sections. The values of the parameters present in the definition are
        always interpreted in regard to their type hint.
        """
        output = []

        for s_name, section in config.items():
            if s_name not in self:
                if s_name == INI_UNIT_SECTION:
                    self[INI_UNIT_SECTION] = section_free_t.NewForUnits()
                else:
                    output.append(f"Section {s_name} does not exist.")
                    continue

            target = self[s_name]
            if s_name == INI_UNIT_SECTION:
                for name_full, value in section.items():
                    if "." in name_full:
                        if name_full not in target:
                            if (
                                should_interpret_values
                                and isinstance(value, str)
                                and not value.isidentifier()
                            ):
                                value = ATOMIC_FACTORIES.NewFromStr(value)
                            target[name_full] = prm_t.NewRuntime(
                                value, target.ppt.basic
                            )
                    else:
                        output.append(
                            f'Name {name_full} does not follow the format '
                            f'"section.parameter".'
                        )
            elif target.__class__.is_controlled:
                for p_name, value in section.items():
                    # The controlling value is not necessarily know yet, so all
                    # controlled parameters with a matching name will be set.
                    parameters = target.ControlledParameters(p_name, with_free=True)
                    if parameters.__len__() > 0:
                        for parameter in parameters:
                            output.extend(parameter.Assign(value))
                    else:
                        output.append(
                            f"Section {s_name} does not contain a parameter {p_name}."
                        )
            else:
                for p_name, value in section.items():
                    if p_name in target:
                        output.extend(target[p_name].Assign(value))
                    elif target.ppt.growable:
                        if p_name.isidentifier():
                            if should_interpret_values and isinstance(value, str):
                                value = ATOMIC_FACTORIES.NewFromStr(value)
                            target[p_name] = prm_t.NewRuntime(value, target.ppt.basic)
                        else:
                            output.append(
                                f"Parameter name {p_name} is not a valid identifier."
                            )
                    else:
                        output.append(
                            f"Section {s_name} does not contain a parameter {p_name}."
                        )

        if INI_UNIT_SECTION in self:
            for name_full, prm in self[INI_UNIT_SECTION].items():
                s_name, p_name = name_full.split(".", maxsplit=1)
                if not ((s_name in self) and (p_name in self[s_name])):
                    output.append(
                        f"No section {s_name} and/or no parameter {p_name} in section."
                    )
                    continue

                target = self[s_name]
                if target.__class__.is_controlled:
                    # All controlled parameters with a matching name will get the
                    # same unit.
                    parameters = target.ControlledParameters(p_name, with_free=True)
                else:
                    parameters = (target[p_name],)
                for parameter in parameters:
                    output.extend(parameter.SetUnit(prm.value))

        if should_update_interface and (self.interface is not None):
            self.interface.UpdateFromNewDict(output)

        return output

    def SyncWithInterface(self) -> None:
        """"""
        assert self.interface is not None
        self.UpdateFromDict(
            self.interface.active_as_str_dict, should_update_interface=False
        )

    def Reset(self) -> None:
        """"""
        self.clear()
        self.path = None
        self.__post_init__()

    def Issues(
        self,
        /,
        *,
        should_raise: bool = False,
        should_be_staged: bool = False,
        can_be_none: bool = False,
    ) -> list[str] | None | int | h.NoReturn:
        """
        should_raise_exception: If False, then returns the list of issues.
        """
        if self.__len__() == 0:
            message = "Empty specification."
            if should_raise:
                raise Exception(message)
            if should_be_staged:
                L.StageIssue(message)
                return 1
            return [message]

        output = []

        for s_name, section in self.items():
            if not section.__class__.is_controlled:
                continue

            c_s_name, c_p_name, _ = section.controller
            if c_s_name == s_name:
                output.append(f"A Section ({s_name}) cannot control itself.")
            elif self[c_s_name].__class__.is_controlled:
                output.append(
                    f'Section "{c_s_name}" of parameter '
                    f'"{c_p_name}", which controls section '
                    f'"{s_name}", is itself controlled'
                )

        if (n_output := output.__len__()) > 0:
            if should_raise:
                raise Exception("\n".join(output))
            if should_be_staged:
                for element in output:
                    L.StageIssue(element)
                return n_output
            return output

        if should_raise:
            return None
        if should_be_staged:
            return 0
        if can_be_none:
            return None
        return []

    def __str__(self) -> str:
        """"""
        output = []

        for name, section in self.items():
            output.append(f"[{name}]")
            output.extend(text.indent(str(section), MARGIN).splitlines())

        return "\n".join(output)

    __repr__ = __str__


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import inspect as nspt
import typing as h
from pathlib import Path as path_t

from babelwidget.backend.generic.path_chooser import path_chooser_t
from babelwidget.api.backend import backend_t
from babelwidget.api.widget import base_h as library_wgt_h
from conf_ini_g.interface.storage.config import SaveConfigToINIDocument
from conf_ini_g.interface.window.config.action import ActionButtonsLayout
from conf_ini_g.interface.window.config.advanced import AdvancedModeLayout
from conf_ini_g.interface.window.config.title import TitleLayout
from conf_ini_g.interface.window.section.collection import SectionsAndCategories
from conf_ini_g.task.formatting import FormattedName
from conf_ini_g.type.config import config_t as functional_t
from conf_ini_g.type.dict import config_str_h, config_typed_h
from logger_36 import L
from value_factory.api.type import invalid_t, unset_t


@d.dataclass(repr=False, eq=False)
class config_t:
    """
    The class cannot use slots because it disables weak referencing, which is required.
    See error message below when using slots:
    TypeError: cannot create weak reference to 'config_t' object
    [...]
    File "[...]conf_ini_g/catalog/interface/window/backend/pyqt5/widget/choices.py", line 41, in SetFunction
        self.clicked.connect(function)
        │                    └ <bound method config_t.ToggleAdvancedMode of <conf_ini_g.interface.window.config.config_t object at [...]>>
        └ <conf_ini_g.catalog.interface.window.backend.pyqt5.widget.choices.radio_choice_wgt_h object at [...]>

    Widget might not cooperate well with list, in which case Python raises the
    following exception: TypeError: multiple bases have instance lay-out conflict
    To be safe, "sections" is a field instead of being part of the class definition.

    _widget: Both an access for interacting with widgets, and a reference keeper to
    prevent autonomous widgets from loosing their "liveness".
    """

    functional: functional_t
    UpdateHistory: h.Callable[[path_t | str], None] | None
    #
    backend: backend_t
    library_wgt: library_wgt_h
    #
    Action: h.Callable[[config_typed_h], None] | None = None
    #
    _widget: dict[str, library_wgt_h] = d.field(init=False, default_factory=dict)

    @property
    def active_as_str_dict(self) -> config_str_h:
        """"""
        return {
            _key: _vle.interface.active_as_str_dict
            for _key, _vle in self.functional.items()
        }

    def __post_init__(self) -> None:
        """"""
        self.functional.interface = self

        if not ((self.functional.path is None) or (self.UpdateHistory is None)):
            self.UpdateHistory(self.functional.path)

    @classmethod
    def New(
        cls,
        title: str | None,
        functional: functional_t,
        backend: backend_t,
        /,
        *,
        history: h.Sequence[str] | None = None,
        UpdateHistory: h.Callable[[path_t | str], None] | None = None,
        action: tuple[str, h.Callable[[config_typed_h], None]] = None,
        advanced_mode: bool = False,
    ) -> h.Self:
        """"""
        if action is None:
            kwargs = {}
        else:
            kwargs = {"Action": action[1]}

        output = cls(
            functional=functional,
            backend=backend,
            library_wgt=backend.base_t(),
            UpdateHistory=UpdateHistory,
            **kwargs,
        )

        path_ini = functional.path

        if path_ini is not None:
            as_str = str(path_ini)
            if history is None:
                history = (as_str,)
            elif as_str not in history:
                history = list(history)
                history.append(as_str)

        # --- Top-level widgets
        (
            title_lyt,
            ini_path_wgt,
            history_button,
            history_menu,
            close_button,
        ) = TitleLayout(
            title,
            functional,
            history,
            backend,
            path_ini,
            functional.UpdateFromINI,
            output.library_wgt.close,
        )
        advanced_mode_lyt, adv_mode_wgt = AdvancedModeLayout(
            advanced_mode, backend, output.ToggleAdvancedMode
        )
        button_lyt, action_button, action_wgt_s = ActionButtonsLayout(
            action,
            path_ini is not None,
            backend,
            output.ShowInConsole,
            output.SaveConfig,
            output.LaunchAction,
            output.library_wgt.close,
        )
        output._widget["path_ini"] = ini_path_wgt
        output._widget["history_button"] = history_button
        output._widget["history_menu"] = history_menu
        output._widget["adv_mode"] = adv_mode_wgt
        output._widget["action"] = action_button
        output._widget["action_buttons"] = action_wgt_s
        output._widget["close"] = close_button

        # --- Sections
        category_selector = SectionsAndCategories(functional, None, backend)
        output._widget["category_selector"] = category_selector

        # --- Layout
        layout = backend.grid_lyt_t()
        if title_lyt is None:
            first_available_row = 0
        else:
            layout.addLayout(title_lyt, 0, 0, 1, 1)
            first_available_row = 1
        layout.addWidget(category_selector, first_available_row, 0, 1, 1)
        layout.addLayout(advanced_mode_lyt, first_available_row + 1, 0, 1, 1)
        layout.addLayout(button_lyt, first_available_row + 2, 0, 1, 1)
        output.library_wgt.setLayout(layout)

        return output

    def ReassignCloseButtonTarget(self) -> None:
        """"""
        current = self.library_wgt
        main_window = None
        while current is not None:
            main_window = current
            current = current.parent()

        self._widget["close"].SetFunction(main_window.close)

    def ToggleAdvancedMode(self, advanced_mode: bool, /) -> None:
        """"""
        for s_name, section in self.functional.items():
            if section.ppt.basic:
                should_check_parameters = True
            elif advanced_mode:
                section.interface.library_wgt.setVisible(True)
                should_check_parameters = True
            else:
                section.interface.library_wgt.setVisible(False)
                should_check_parameters = False

            if should_check_parameters:
                for _, parameter in section.active_parameters:
                    if not parameter.ppt.basic:
                        if advanced_mode:
                            parameter.interface.SetVisible(True)
                        else:
                            parameter.interface.SetVisible(False)

    def SyncWithFunctional(self) -> None:
        """"""
        config_typed = self.functional.active_as_typed_dict

        for s_name, section in self.functional.items():
            section_typed = config_typed[s_name]

            for p_name, parameter in section.prm_iterator:
                interface = parameter.interface
                if p_name in section_typed:
                    value = section_typed[p_name]
                    if not isinstance(value, invalid_t | unset_t):
                        interface.value.Assign(value, parameter.hint)
                        unit = parameter.unit_name
                        if unit is not None:
                            interface.unit.setText(unit)
                elif parameter.ppt.optional:
                    interface.value.Assign(parameter.default, parameter.hint)

    def UpdateFromNewDict(self, issues: h.Sequence[str], /) -> None:
        """"""
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        category_selector = self._widget["category_selector"]
        if isinstance(category_selector, self.backend.tabs_t):
            # Note: idx^th layout: category_selector.widget(t_idx).widget().layout().
            while category_selector.count() > 0:
                category_selector.removeTab(0)
        else:
            layout = category_selector.widget().layout()
            while layout.count() > 0:
                layout.itemAt(0).widget().setParent(None)

        # TODO: Why is it necessary to call SectionsAndCategories again? Actually, the
        #     problem could be more general: How to handle parameters present in the new
        #     dict/INI file that were not present in the previous/initial INI file?
        _ = SectionsAndCategories(self.functional, category_selector, self.backend)
        self.SyncWithFunctional()
        self.ToggleAdvancedMode(self._widget["adv_mode"].true_btn.isChecked())

        path_ini = self.functional.path
        self._widget["path_ini"].Assign(path_ini, None)
        self._widget["history_button"].setEnabled(True)
        if str(path_ini) not in (
            _elm.text() for _elm in self._widget["history_menu"].actions()
        ):
            self._widget["history_menu"].addAction(str(path_ini))

        if self.UpdateHistory is not None:
            self.UpdateHistory(path_ini)

    def LaunchAction(self) -> None:
        """"""
        self.setEnabled(False)
        self.backend.qt_core_app_t.processEvents()

        try:
            self.functional.SyncWithInterface()
            self.Action(self.functional.active_as_typed_dict)
        except Exception as exception:
            trace = nspt.trace()[-1]
            context = "\n".join(trace.code_context)
            self.backend.ShowErrorMessage(
                f"{trace.filename}@{trace.lineno}:{trace.function}\n"
                f"{context}\n"
                f"{exception}",
                parent=self.library_wgt,
            )

        self.setEnabled(True)

    def ShowInConsole(self) -> None:
        """"""
        as_str = []
        max_length = -1
        for s_name, section in self.active_as_str_dict.items():
            as_str.append(f"{FormattedName(s_name).upper()}")
            for p_name, parameter in section.items():
                formatted = FormattedName(p_name)
                as_str.append(f"    {formatted}:{parameter}")
                max_length = max(max_length, formatted.__len__() + 4)

        as_str = "\n".join(map(lambda _: _AlignSemiColon(_, max_length), as_str))

        L.info(f"CONFIGURATION\n{as_str}")

    def SaveConfig(self, new_ini: bool, /) -> None:
        """"""
        if new_ini:
            path_chooser = path_chooser_t.New("output", "document", message="Save Config As")
            path = path_chooser.NewSelected()
            if path is None:
                return

            self.functional.path = path
        elif self.functional.path is None:
            return

        issues = SaveConfigToINIDocument(self.active_as_str_dict, self.functional.path)
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self)

    def __getattr__(self, attribute: str, /) -> h.Any:
        """
        E.g., used for "show".
        """
        try:
            output = object.__getattribute__(self, attribute)
        except AttributeError:
            output = getattr(self.library_wgt, attribute)

        return output


def _AlignSemiColon(string: str, max_length: int, /) -> str:
    """"""
    if ":" not in string:
        return string

    name, value = string.split(":")
    return f"{name}: {(max_length - name.__len__()) * ' '}{value}"


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import typing as h

from babelwidget.api.backend import backend_t
from babelwidget.api.widget import base_h as library_wgt_h
from babelwidget.api.widget import dropdown_choice_h as dropdown_choice_wgt_h
from babelwidget.api.widget import text_line_h as text_line_wgt_h
from conf_ini_g.interface.window.parameter.value import value_wgt_a
from value_factory.api.catalog import collection_t
from value_factory.api.type import hint_t

_DELIMITERS_TYPES = (tuple, list, set)
_DELIMITERS_VALUES = ("()", "[]", "{}")
_DELIMITERS = dict(zip(_DELIMITERS_TYPES, _DELIMITERS_VALUES))
_ANY_LENGTH = "Any"


@d.dataclass(repr=False, eq=False)
class collection_wgt_t(value_wgt_a):
    """
    Cannot use slots (weak reference issue).
    """

    ENTRY_ANY: h.ClassVar[str] = "any"
    ENTRIES: h.ClassVar[tuple[str, ...]] = ("1", "2", "3", "4", "5", "6", ENTRY_ANY)

    library_wgt: library_wgt_h
    delimiters: h.Literal["()", "[]", "{}"] = d.field(
        init=False, default=_DELIMITERS[tuple]
    )
    length_selector: dropdown_choice_wgt_h = d.field(init=False, default=None)
    components: tuple[text_line_wgt_h, ...] = d.field(init=False, default=None)

    @classmethod
    def New(
        cls,
        stripe: hint_t | None,
        backend: backend_t,
        /,
    ) -> h.Self:
        """
        If stripe does not contain the necessary details, a simple free-text input
        widget is used. If the value is not coherent with the details (which should not
        happen if stripe contains the necessary details and the value has been
        validated), a choice with the length of the value is added, with an exclamation
        point.
        """
        components, length_selector = cls._SetUpComponents(stripe, backend)
        output = cls(components, "textChanged", backend, library_wgt=backend.base_t())
        length_selector.SetFunction(output._AccountForLengthSwitch)
        output.components = components
        output.length_selector = length_selector

        layout = backend.hbox_lyt_t()
        layout.setContentsMargins(0, 0, 0, 0)
        if output.length_selector is not None:
            layout.addWidget(output.length_selector)
        for component in output.components:
            layout.addWidget(component)
        output.library_wgt.setLayout(layout)

        return output

    def _AccountForLengthSwitch(self, new_index: int, /) -> None:
        """"""
        new_length = self.length_selector.ItemAt(new_index)
        if new_length == self.__class__.ENTRY_ANY:
            new_length = 1
        elif new_length.endswith("!"):
            new_length = int(new_length[:-1])
        else:
            new_length = int(new_length)
        _AdjustComponentVisibility(self.components, new_length)

    @classmethod
    def _SetUpComponents(
        cls, stripe: hint_t | None, backend: backend_t, /
    ) -> tuple[tuple[text_line_wgt_h, ...], dropdown_choice_wgt_h | None]:
        """"""
        if isinstance(stripe, hint_t):
            annotation = stripe.FirstAnnotationWithType(collection_t)
        else:
            annotation = None
        if (annotation is None) or annotation.accepts_any_length:
            entries = cls.ENTRIES
            max_entry = int(cls.ENTRIES[-2])
        else:
            lengths = annotation.lengths
            entries = tuple(map(str, lengths))
            max_entry = lengths[-1]

        if max_entry is None:
            components = (backend.text_line_t(parent=None),)
        else:
            components = tuple(
                backend.text_line_t(parent=None) for _ in range(max_entry)
            )

        if entries.__len__() > 1:
            _AdjustComponentVisibility(components, 1)

            length_selector = backend.dropdown_choice_t()
            for entry in entries:
                length_selector.addItem(entry)
            length_selector.setCurrentText(cls.ENTRY_ANY)
        else:
            length_selector = None

        return components, length_selector

    def Assign(self, value: tuple | list | set, stripe: hint_t | None, /) -> None:
        """"""
        length = value.__len__()

        delimiters = _DELIMITERS.get(type(value), None)
        if delimiters is None:
            raise TypeError(f"{type(value).__name__}: Unhandled sequence type.")
        self.delimiters = delimiters

        if (n_components := self.components.__len__()) == 1:
            value_as_str = str(value)[1:-1] if length > 0 else ""
            if length == 1:
                value_as_str = value_as_str[:-1]
            self.components[0].setText(value_as_str)
        elif length > n_components:
            if (self.length_selector is not None) and (
                self.length_selector.itemText(self.length_selector.count() - 1)
                == _ANY_LENGTH
            ):
                value_as_str = str(value)[1:-1] if length > 0 else ""
                self.components[0].setText(value_as_str)
                self.length_selector.setCurrentText(_ANY_LENGTH)
            else:
                raise ValueError(
                    f"Value {value} has too many elements "
                    f"for widget with at most {n_components} component(s)."
                )
        else:
            for widget, element in zip(self.components[:length], value):
                widget.setText(str(element))
            _AdjustComponentVisibility(self.components, length)
            if self.length_selector is not None:
                self.length_selector.setCurrentText(str(length))

    def Text(self) -> str:
        """"""
        contents = []
        for component in self.components:
            # Do not use "visible" here since setting visible does not really set the
            # property until it is actually shown. The documentation explains about
            # ancestors being visible or not, but it was not clear that the property is
            # apparently not effective immediately.
            if not component.isEnabled():
                break
            component_as_str = component.Text()
            if component_as_str == "":
                break
            contents.append(f'"{component_as_str}"')

        if contents.__len__() == 0:
            return self.delimiters
        elif contents.__len__() == 1:
            if self.delimiters[0] == "(":
                return f"({contents[0]},)"
            else:
                return f"[{contents[0]}]"
        else:
            return self.delimiters[0] + ", ".join(contents) + self.delimiters[1]


def _AdjustComponentVisibility(
    components: h.Sequence[text_line_wgt_h], length: int, /
) -> None:
    """"""
    for c_idx, component in enumerate(components):
        if c_idx < length:
            component.setVisible(True)
            component.setEnabled(True)
        else:
            component.setVisible(False)
            component.setEnabled(False)


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import typing as h
from pathlib import Path as pl_path_t

from babelwidget.api.backend import backend_t
from babelwidget.api.widget import base_h as library_wgt_h
from babelwidget.api.widget import label_h as label_wgt_h
from babelwidget.api.widget import path_chooser_t
from babelwidget.api.widget import text_line_h as text_line_wgt_h
from conf_ini_g.constant.window import PATH_SELECTOR_WIDTH
from conf_ini_g.interface.window.parameter.value import value_wgt_a
from value_factory.api.catalog import path_kind_e, path_purpose_e, path_t
from value_factory.api.type import hint_t


@d.dataclass(repr=False, eq=False)
class path_wgt_t(value_wgt_a):
    """
    Cannot use slots (weak reference issue).
    """

    PostAssignmentFunction: h.Callable[[pl_path_t], None] | None
    library_wgt: library_wgt_h
    editable: bool
    backend_for_selection: backend_t
    type_: path_kind_e | None = d.field(init=False, default=None)
    path: text_line_wgt_h | label_wgt_h | None = d.field(init=False, default=None)
    _path_chooser: path_chooser_t | None = d.field(
        init=False, default=None
    )

    @classmethod
    def New(
        cls,
        stripe: hint_t,
        backend: backend_t,
        /,
        *,
        editable: bool = True,
        PostAssignmentFunction: h.Callable[[pl_path_t], None] | None = None,
    ) -> h.Self:
        """
        If stripe does not contain the necessary details, the target type is set to any and considered as input, and
        the selection button label ends with an exclamation point.
        """
        if editable:
            path = backend.text_line_t()
            messenger = path
        else:
            path = backend.label_t()
            messenger = None
        output = cls(
            messenger,
            "textChanged",
            backend,
            PostAssignmentFunction=PostAssignmentFunction,
            library_wgt=backend.base_t(),
            editable=editable,
            backend_for_selection=backend,
        )
        output.path = path

        annotation = stripe.FirstAnnotationWithType(path_t)
        if annotation is None:
            path_purpose = path_purpose_e.input
            path_kind = path_kind_e.any
        else:
            path_purpose = annotation.purpose
            path_kind = annotation.kind

        output.type_ = path_kind
        output._path_chooser = path_chooser_t.New(path_purpose.name, path_kind.name)

        if path_kind is path_kind_e.document:
            selector_label = "🗋"
        elif path_kind is path_kind_e.folder:
            selector_label = "📂"
        else:
            selector_label = "📂🗋"
        if path_purpose is path_purpose_e.input:
            selector_color = "green"
        elif path_purpose is path_purpose_e.output:
            selector_color = "red"
        else:
            selector_color = "blue"
        path_selector = backend.button_t(selector_label, parent=output.library_wgt)
        path_selector.SetFunction(output.SelectDocument)

        path_selector.setStyleSheet(f"color: {selector_color};")
        path_selector.setFixedWidth(PATH_SELECTOR_WIDTH)

        layout = backend.hbox_lyt_t()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(path)
        layout.addWidget(path_selector)
        output.library_wgt.setLayout(layout)

        return output

    def Assign(self, value: pl_path_t, _: h.Any, /) -> None:
        """"""
        value = _ValidStrValue(value, self.editable)
        self.path.setText(value)

    def Text(self) -> str | None:
        """
        /!\\ pathlib.Path("") == pathlib.Path(".").
        """
        return self.path.Text()

    def SelectDocument(self) -> None:
        """"""
        current_path = self.Text()
        current_doc = pl_path_t(current_path).resolve()

        # USed to use self.backend_for_selection.
        self._path_chooser.SetFolder(current_doc.parent)
        selection = self._path_chooser.NewSelected()
        if selection is None:
            return

        self.Assign(selection, None)
        # Put post-assignment call here instead of in the Assign method in case Assign
        # is also called at initialization time, in New, one day. Indeed, the
        # post-assignment task should have been done already, or will be done, when
        # instantiating the interface.
        if self.PostAssignmentFunction is not None:
            self.PostAssignmentFunction(selection)


def _ValidStrValue(value: pl_path_t, editable: bool, /) -> str:
    """"""
    if not editable:
        try:
            value = value.resolve(strict=True).relative_to(
                pl_path_t.home().resolve(strict=True)
            )
        except ValueError:
            # On Linux, this happens when home is a bind mount.
            pass

    return str(value)


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import textwrap as text
import typing as h

from conf_ini_g.constant.console import MARGIN
from conf_ini_g.type.dict import parameter_name_h, section_name_h, section_typed_h
from conf_ini_g.type.parameter import ctl_t, prm_t
from conf_ini_g.type.property import ppt_t
from logger_36 import L
from value_factory.api.constant import UNSET
from value_factory.api.type import unset_t

section_ppty_h = str
controlling_value_h = str

section_definition_h = dict[
    section_ppty_h | parameter_name_h | controlling_value_h,
    ppt_t | prm_t | dict[parameter_name_h, prm_t],
]


@d.dataclass(slots=True, repr=False, eq=False)
class section_free_t(dict[parameter_name_h, prm_t]):
    is_controlled: h.ClassVar[bool] = False  # To avoid having to use isinstance.

    ppt: ppt_t | None = None
    interface: h.Any = None

    @property
    def prm_iterator(self) -> h.Iterator[tuple[str, prm_t]]:
        """"""
        return iter(self.items())

    @property
    def active_parameters(self) -> tuple[[str, prm_t], ...]:
        """"""
        return tuple(self.items())

    @property
    def active_as_typed_dict(self) -> section_typed_h:
        """"""
        return {_key: _vle.value for _key, _vle in self.items()}

    def __post_init__(self) -> None:
        """"""
        if self.ppt is None:
            self.ppt = ppt_t()

    @classmethod
    def New(cls, definition: section_definition_h, /) -> h.Self:
        """"""
        output = cls()

        for name, content in definition.items():
            if isinstance(content, ppt_t):
                output.ppt = content
            else:
                if isinstance(content, prm_t):
                    assert name.isidentifier(), name
                else:
                    assert isinstance(name, str), name
                output[name] = content

        if output.ppt.growable is None:
            output.ppt.growable = output.__len__() == 0

        # Wait for the section to be fully initialized (optional plugin additions)
        # before checking if it has issues.

        return output

    @classmethod
    def NewForUnits(cls) -> h.Self:
        """"""
        output = cls()

        properties = output.ppt
        properties.short = "Parameter units"
        properties.long = (
            "Section to specify parameter units is the form "
            '"section.parameter = <SciPy_unit>" '
            "where <SciPy_unit> is a unit or SI prefix defined by the"
            "SciPy library. "
            "See: https://docs.scipy.org/doc/scipy/reference/constants.html"
        )
        properties.basic = True
        properties.growable = True

        return output

    def Issues(
        self,
        /,
        *,
        should_raise: bool = False,
        should_be_staged: bool = False,
        can_be_none: bool = False,
    ) -> list[str] | None | int | h.NoReturn:
        """"""
        return _Issues(self, should_raise, should_be_staged, can_be_none)

    def __str__(self) -> str:
        """"""
        return _AsStr(self)

    __repr__ = __str__


@d.dataclass(slots=True, repr=False, eq=False)
class section_controlled_t(
    dict[controlling_value_h | None, dict[parameter_name_h, prm_t]]
):
    """
    Free parameter: As opposed to controlled.

    There is an implicit constraint that if two parameters in two controlled subsets
    share the same name, they must have the same hint and unit.

    controller:
    - UNSET:                 The section is not functional yet,
    - s_name, p_name, ctl_t: The section is functional.
    """

    is_controlled: h.ClassVar[bool] = True  # To avoid having to use isinstance.

    ppt: ppt_t | None = None
    controller: tuple[section_name_h, parameter_name_h, ctl_t] | unset_t = d.field(
        init=False, default=UNSET
    )
    interface: h.Any = None

    @property
    def controlling_value(self) -> str:
        """"""
        return self.controller[2].value

    @property
    def controlling_values(self) -> tuple[str, ...] | unset_t:
        """"""
        if self.__len__() > 0:
            return tuple(sorted(_ for _ in self.keys() if _ is not None))
        return UNSET

    @property
    def prm_free_names(self) -> tuple[str, ...]:
        """"""
        if None in self:
            return tuple(self[None].keys())
        return ()

    @property
    def prm_iterator(self) -> h.Iterator[tuple[str, prm_t]]:
        """"""
        for subset in self.values():
            for name, parameter in subset.items():
                yield name, parameter

    @property
    def active_parameters(self) -> tuple[[str, prm_t], ...]:
        """"""
        if None in self:
            output = list(self[None].items())
        else:
            output = []
        output.extend(self[self.controlling_value].items())

        return tuple(output)

    @property
    def active_as_typed_dict(self) -> section_typed_h:
        """"""
        if None in self:
            output = {_key: _vle.value for _key, _vle in self[None].items()}
        else:
            output = {}
        output.update(
            {_key: _vle.value for _key, _vle in self[self.controlling_value].items()}
        )

        return output

    @property
    def prm_iterator_per_controlling_value(
        self,
    ) -> h.Iterator[tuple[str, tuple[tuple[str, prm_t], ...]]]:
        """"""
        if None in self:
            free_parameters = tuple(self[None].items())
        else:
            free_parameters = ()
        for value in self.controlling_values:
            yield value, free_parameters + tuple(self[value].items())

    def __post_init__(self) -> None:
        """"""
        if self.ppt is None:
            self.ppt = ppt_t()

    @classmethod
    def New(cls, definition: section_definition_h, /) -> h.Self:
        """"""
        output = cls()

        is_empty = definition.__len__() == 0
        is_empty_w_ppt = (definition.__len__() == 1) and isinstance(
            definition[tuple(definition.keys())[0]], ppt_t
        )
        # /!\\ ppt_t must not be a dict subclass, or has_dict_s becomes invalid.
        has_dict_s = any(isinstance(_, dict) for _ in definition.values())
        assert is_empty or is_empty_w_ppt or has_dict_s, (
            is_empty,
            is_empty_w_ppt,
            has_dict_s,
        )

        for name, content in definition.items():
            if isinstance(content, ppt_t):
                output.ppt = content
            elif isinstance(content, prm_t):
                assert name.isidentifier(), name
                if None in output:
                    output[None][name] = content
                else:
                    output[None] = {name: content}
            else:
                assert isinstance(name, str), name
                output[name] = content

        if output.ppt.growable is None:
            output.ppt.growable = output.__len__() == 0

        # Wait for the section to be fully initialized (optional plugin additions)
        # before checking if it has issues.

        return output

    def Issues(
        self,
        /,
        *,
        should_raise: bool = False,
        should_be_staged: bool = False,
        can_be_none: bool = False,
    ) -> list[str] | None | int | h.NoReturn:
        """"""
        return _Issues(self, should_raise, should_be_staged, can_be_none)

    def ControlledParameters(
        self, name: str | None, /, *, with_free: bool = False
    ) -> tuple[prm_t, ...] | tuple[tuple[str, prm_t], ...]:
        """
        name: None=all.
        """
        output = []

        values = self.controlling_values
        if with_free and (None in self):
            values += (None,)
        for value in values:
            subset = self[value]
            if name is None:
                output.extend(subset.items())
            elif name in subset:
                output.append(subset[name])

        return tuple(output)

    def __str__(self) -> str:
        """"""
        return _AsStr(self)

    __repr__ = __str__


def _Issues(
    section: section_free_t | section_controlled_t,
    should_raise: bool,
    should_be_staged: bool,
    can_be_none: bool,
    /,
) -> list[str] | None | int | h.NoReturn:
    """"""
    output = []

    basic = section.ppt.basic
    optional = section.ppt.optional
    growable = section.ppt.growable

    if not (section.ppt.basic or section.ppt.optional):
        output.append("Section is not basic but not optional.")

    if (section.__len__() == 0) and not growable:
        output.append("Empty non-growable section")

    if section.__len__() > 0:
        if section.__class__.is_controlled:
            # This might change if adding a runtime parameter can be done while
            # specifying a controlling value. However, adding a controlled parameter
            # will require checking that is does not have the name of a free one.
            if growable:
                output.append("A Controlled section cannot be growable.")

            if None in section:
                free = set(section[None].keys())
                if any(
                    free.intersection(section[_].keys()).__len__() > 0
                    for _ in section.controlling_values
                ):
                    output.append(
                        "Free parameter names cannot appear in the controlled subsets."
                    )

            for p_name, parameter in section.ControlledParameters(None):
                if not parameter.ppt.optional:
                    output.append(f'Controlled parameter "{p_name}" must be optional.')

            for p_name, parameter in section.prm_iterator:
                # Forbid controlling cascade to avoid cycles.
                if isinstance(parameter, ctl_t):
                    output.append(
                        f"Parameter {p_name} in controlled section "
                        f"cannot be itself a controller."
                    )

        n_parameters = 0
        n_basic_prms = 0
        for p_name, parameter in section.prm_iterator:
            n_parameters += 1
            if parameter.ppt.basic:
                n_basic_prms += 1

            if parameter.ppt.basic and not basic:
                output.append(f'Basic parameter "{p_name}" in advanced section')
            if optional and not parameter.ppt.optional:
                output.append(f'Mandatory parameter "{p_name}" in optional section')

        if basic and (n_parameters > 0) and (n_basic_prms == 0):
            output.append("Basic section without any basic parameters")

    if (n_output := output.__len__()) > 0:
        if should_raise:
            raise Exception("\n".join(output))
        if should_be_staged:
            for element in output:
                L.StageIssue(element)
            return n_output
        return output

    if should_raise:
        return None
    if should_be_staged:
        return 0
    if can_be_none:
        return None
    return []


def _AsStr(section: section_free_t | section_controlled_t, /) -> str:
    """"""
    if section.ppt is None:
        # This should only happen in the recursive call triggered below.
        output = []
    else:
        output = ["Properties"]
        output.extend(text.indent(str(section.ppt), MARGIN).splitlines())
        if section.__class__.is_controlled:
            output.append(f"{MARGIN}Control: {section.controlling_value}")

    for name, content in section.items():
        if isinstance(content, prm_t):
            output.append(f"* {name}")
            output.extend(text.indent(str(content), MARGIN).splitlines())
        else:
            output.append(f"<= {name}")
            sub_section = section_free_t.New(content)
            sub_section.ppt = None  # /!\\ For convenient re-use of __str__.
            output.extend(text.indent(str(sub_section), MARGIN).splitlines())

    return "\n".join(output)


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""



"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2021
SEE COPYRIGHT NOTICE BELOW
"""

import re as rgex
import typing as h
from argparse import ArgumentParser as argument_parser_t
from pathlib import Path as path_t

from conf_ini_g.extension.string import Flattened
from conf_ini_g.type.config import config_t
from conf_ini_g.type.dict import config_typed_h
from value_factory.api.constant import UNSET
from value_factory.api.type import unset_t

parsed_arguments_h = dict[str, str | unset_t]


# Specified INI document file is stored in _INI_DOCUMENT_VARIABLE.
_INI_DOCUMENT_VARIABLE = "ini_path"

_ADVANCED_MODE_OPTION = "advanced-mode"
_ADVANCED_MODE_VARIABLE = "advanced_mode"

# Usage: {section_name}{_SECTION_PARAMETER_SEPARATOR}{parameter_name}.
_SECTION_PARAMETER_SEPARATOR = "-"


def CommandLineParser(
    description: str | None, specification: config_t, /
) -> argument_parser_t:
    """
    Currently, controlled parameters cannot be passed on the command line.
    """
    output = argument_parser_t(description=description, allow_abbrev=False)
    # This argument is optional because all parameters can be explicitly passed as
    # arguments (CLI mode), or nothing can be passed at all (GUI mode).
    output.add_argument(
        dest=_INI_DOCUMENT_VARIABLE,
        help="Path to INI configuration file",
        default=None,
        nargs="?",
        metavar="INI_config_file",
    )

    # Default is an unset_t instance to avoid overwriting if in INI but not passed.
    default = UNSET
    options = []
    for s_name, section in specification.items():
        if section.__class__.is_controlled:
            if None in section:
                parameters = section[None]
            else:
                # Controlled parameters are skipped.
                continue
        else:
            parameters = section

        for p_name, parameter in parameters.items():
            option = f"{s_name}{_SECTION_PARAMETER_SEPARATOR}{p_name}"
            if option == _ADVANCED_MODE_OPTION:
                # Raising an exception is adapted here since execution has been launched
                # from command line.
                raise ValueError(
                    f"{option}: Command-line option for parameter is identical to "
                    f"advanced mode option. "
                    f"Please change parameter specification (section name and/or "
                    f"parameter name)."
                )
            if option in options:
                # This can happen in controlled sections whose alternatives share
                # equally named parameters.
                continue

            attribute = f"{s_name}{_SECTION_PARAMETER_SEPARATOR}{p_name}"

            if parameter.ppt.optional:
                if isinstance(parameter.default, str):
                    delimiter = '"'
                else:
                    delimiter = ""
                type_and_value = (
                    f"Type: {parameter.hint}. "
                    f"Default: [green]{delimiter}{parameter.default}{delimiter}[/]"
                )
            else:
                type_and_value = str(default)
            flattened = Flattened(type_and_value)
            definition = f"{parameter.ppt.short}. {flattened}"

            # Type could be TypeOfAnnotatedHint(cmd_line_type). However, to allow
            # passing any of the allowed types, deferring type validation to functional
            # config instantiation, this parameter is not passed.
            output.add_argument(
                f"--{option}",
                dest=attribute,
                help=definition,
                default=default,
                metavar=attribute,
            )
            options.append(option)

    output.add_argument(
        f"--{_ADVANCED_MODE_OPTION}",
        dest=_ADVANCED_MODE_VARIABLE,
        help="Toggle display of advanced sections and parameters",
        action="store_true",
    )

    return output


def CommandLineConfig(
    parser: argument_parser_t, /, *, arguments: h.Sequence[str] = None
) -> tuple[config_typed_h, bool, path_t | None]:
    """"""
    parsed, unknowns = parser.parse_known_args(args=arguments)
    parsed = vars(parsed)

    advanced_mode = parsed[_ADVANCED_MODE_VARIABLE]
    del parsed[_ADVANCED_MODE_VARIABLE]

    ini_path = parsed[_INI_DOCUMENT_VARIABLE]
    del parsed[_INI_DOCUMENT_VARIABLE]
    if ini_path is not None:
        try:
            ini_path = path_t(ini_path).resolve(strict=True)
        except (FileNotFoundError, RuntimeError):
            raise FileNotFoundError(f"Nonexistent or invalid path: {ini_path}")

    pattern = rgex.compile(
        r"--(\w+)" + _SECTION_PARAMETER_SEPARATOR + r"(\w+\.\w+)=(.+)", flags=rgex.ASCII
    )
    for unknown in unknowns:
        match = pattern.fullmatch(unknown)
        if match is None:
            # Raising an exception is adapted here since execution has been launched
            # from command line.
            raise ValueError(
                f"{unknown}: Invalid option syntax; Expected=--SECTION-PARAMETER=VALUE"
            )

        section, parameter, value = match.groups()
        parsed[f"{section}{_SECTION_PARAMETER_SEPARATOR}{parameter}"] = value

    raw_config = {}
    for sct_name, prm_name, value in _SectionParameterValueIterator(parsed):
        if sct_name in raw_config:
            raw_config[sct_name][prm_name] = value
        else:
            raw_config[sct_name] = {prm_name: value}

    return raw_config, advanced_mode, ini_path


def _SectionParameterValueIterator(
    arguments: parsed_arguments_h,
) -> h.Iterator[tuple[str, str, h.Any]]:
    """"""
    for prm_uid, value in arguments.items():
        # See CommandLineParser for why this can happen.
        if value is UNSET:
            continue

        section, parameter = prm_uid.split(_SECTION_PARAMETER_SEPARATOR)
        yield section, parameter, value


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""
